##################################################################################
#
#   Kaivo Public Software License
#
#   Copyright (c) 2001, Kaivo, Inc.
#   All rights reserved.
#
#   Redistribution and use in source and binary forms, with or without
#   modification, are permitted provided that the following conditions are met:
#
#       o Redistributions of source code must retain the above copyright notice,
#         this list of conditions and the following disclaimer.
#
#       o Redistributions in binary form must reproduce the above copyright
#         notice, this list of conditions and the following disclaimer
#         in the documentation and/or other materials provided with the
#         distribution.
#
#       o Neither Kaivo nor the names of its contributors may be used to endorse
#         or promote products derived from this software without specific prior
#         written permission.
#
#   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#   AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
#   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#   ARE DISCLAIMED. IN NO EVENT SHALL KAIVO OR ITS CONTRIBUTORS BE LIABLE FOR 
#   ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
#   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
#   LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND 
#   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
#   (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
#   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
##################################################################################

__doc__ = """Document Library Document"""
__version__ = '1.0b2'

# Document by Casey Duncan (cduncan@kaivo.com)
# Kaivo, Inc. (http://www.kaivo.com)

from string import split, join, strip, find, rfind, replace
from time import time
from os import path, listdir
from tempfile import mktemp
import urllib
from StringIO import StringIO
from Globals import DTMLFile, package_home
from DateTime import DateTime
from OFS.Image import *
from OFS.PropertySheets import FixedSchema
from Products.ZCatalog.CatalogAwareness import CatalogAware
from App.ImageFile import ImageFile
import DLGlobals
import Acquisition
from FileConverters import FileConverters

manage_addDocumentFileForm = DTMLFile('dtml/AddDocumentForm', globals())

def check_content_type(context, content_type):
    """Verify that the type of file is allowed"""
 
    if hasattr(context, 'deny_content_types') and content_type:
        return not (content_type in context.deny_content_types)
    
    return 1

def manage_addDocumentFile(self,id='',file='',title='',precondition='', content_type='',
                   REQUEST=None):
    """Creates a new Library Document File"""

    id=str(id)
    title=str(title)
    content_type=str(content_type)
    precondition=str(precondition)

    self=self.this()
    doc = Document(id, title, '', content_type, precondition, container=self)
    self._setObject(doc.getId(), doc)
    doc = self._getOb(doc.getId())
    # Now we "upload" the data.  By doing this in two steps, we
    # can use a database trick to make the upload more efficient.
    doc.manage_upload(file)

    if content_type:
        doc.content_type=content_type

    if not check_content_type(self, doc.content_type):
        doc.manage_delObjects(doc.getId())
        raise "IllegalFileType", \
              "<span>That file type cannot be uploaded into the library.</span>"

    if REQUEST is not None:
        REQUEST['RESPONSE'].redirect(self.absolute_url()+'/manage_main')
        
class Document(CatalogAware, File):
    """Document library document"""

    meta_type = 'Document File'

    __ac_permissions__ = File.__ac_permissions__ + (
        ('Access contents information', ('format', 'source', 'identifier', 'subject',
                                         'content_description', 'content_size', 'is_accepted',
                                         'owner_id', 'authenticated_user_is_owner')),
        ('View', ('text_content', 'PrincipiaSearchSource', 'read', 'super_topics',
                                  'topicMap', 'all_searchable_text')),
        ('Manage properties', ('acceptDocument', 'rejectDocument', 'setIcons', 'showTopics')),
    )

    manage_options = File.manage_options + (
        { 'label': 'View Plain Text', 'action': 'text_content' },
    )

    _properties = ()

    # Set default values for important properties
    review_date = DLGlobals.document_unreviewed_date

    def __init__(self, id='', title='', file='', content_type='', precondition='', container=None):
        if not check_content_type(self, content_type):
            raise "IllegalFileType", \
                  "<span>That file type cannot be uploaded into the library.</span>"

        if not id and container:
            # Create a randomized id if none was provided
            container = getattr(container, 'aq_base', container)
            while not id or hasattr(container, id):
                id = str(time())

        File.__init__(self, id, title, file, content_type, precondition)

        if not hasattr(self, 'icon'): # We aren't an acquisition wrapper yet
            self.setIcons()

    def __getattr__(self, name):
        """Allow access to index_html via the filename and 
           default property values"""

        if name == self.__dict__.get('filename', None):
            return self.index_html
        elif name[0] != '_' and name[:3] != 'aq_' and self.hasProperty(name):
            # Make default property values act like attributes
            for p in self._properties:
                if p['id'] == name: return p['default']
            raise AttributeError, name
        else:
            raise AttributeError, name

    def acceptDocument(self, REQUEST=None, RESPONSE=None):
        """Mark the document as accepted by setting its review_date to the current date"""
        self.review_date = DateTime()
        self.reindex_object()
        self.showTopics()

        if REQUEST and RESPONSE and REQUEST.has_key('last_url'):
            RESPONSE.redirect(REQUEST.last_url + '?' + REQUEST.get('last_query',''))

    def rejectDocument(self, REQUEST=None, RESPONSE=None):
        """Mark the document as rejected by resetting its review_date"""
        self.review_date = DLGlobals.document_rejected_date
        self.reindex_object()

        if REQUEST and RESPONSE and REQUEST.has_key('last_url'):
            RESPONSE.redirect(REQUEST.last_url + '?' + REQUEST.get('last_query',''))

    def is_accepted(self):
        """Returns true if the document is indexed and has been reviewed and accepted"""
        return hasattr(self.aq_base, 'topics') and self.topics \
               and self.getProperty('review_date') != DLGlobals.document_rejected_date

    #
    #   Content and conversion handling methods
    #

    def read(self):
        """Return raw data of document file
           This can be memory intensive for large files. 
           Use the _writeToTempFile method to get the data to external programs"""
        data=self.data
        if type(data) is type(''): return data

        r = StringIO()
        while data is not None:
            r.write(data.data)
            data=data.next
        
        return r.getvalue()

    def _writeToTempFile(self):
        """Writes the contents of the document file to a temporary file and
           returns the file name. Used by file converters"""
        tmp_file = mktemp()
        f = open(tmp_file, 'wb')

        data=self.data
        if type(data) is type(''): 
            f.write(data)
        else:
            while data is not None:
                f.write(data.data)
                data=data.next

        f.close()
        return tmp_file
        
    def PrincipiaSearchSource(self):
        """Return the raw text representation of the document if possible"""

        # Check for a cached copy of the full text first
        if hasattr(self, '_v_raw_text'):
            return self._v_raw_text

        format = self.format()
        
        if FileConverters.has_key(format):
            # use the installed converter
            convert = FileConverters[format].convert
            try:
                # Convert the document data to text and cache the result
                self._v_raw_text = convert(self)
                return self._v_raw_text
            except:
                return ''
        elif format[:5] == 'text/':
            # This appears to be a raw text type, so return the data as is
            return self.read()
        else:
            # Can't do it captain
            return ''

    def text_content(self, RESPONSE=None):
        """The file converted to text with proper HTTP headers set
           for viewing in a browser"""
        t = self.PrincipiaSearchSource()

        if RESPONSE is not None:
            RESPONSE.setHeader('Last-Modified', rfc1123_date(self._p_mtime))
            RESPONSE.setHeader('Content-Type', 'text/plain')
            RESPONSE.setHeader('Content-Length', len(t))
        
        return t

    def content_description(self):
        """Returns a friendly description of the document content"""
    
        format = self.format()

        if format[:5] == 'text/':
            d = 'Text Document (%s)' % format[5:]
        else:
            sep = find(format,'/') + 1
            d = 'Document (%s)' % format[sep:]
        
        if FileConverters.has_key(format):
            return getattr(FileConverters[format], 'content_description', d)
        else:
            return d

    def content_size(self):
        """Returns a qualified document size in Kb or Mb"""
        s = self.get_size()

        if s == 0:
            return '0 bytes'
        elif s <= 1024:
            return '1 Kb'
        elif s < 1048576:
            return '%d Kb' % (s / 1024)
        else:
            return '%d Mb' % (s / 1048576)
    
    def topicMap(self):
        """Returns a tuple of id,title mappings for each topic this
           document belongs to"""

        r = []
        
        if hasattr(self, 'topics') and hasattr(self.aq_parent, 'Index'):
            indexQuery = self.aq_parent.Index.query # Acquire the index's query method

            # Iterate the topic ids assigned to this document and
            # search the topic index to find the titles of the topics
            for topic_id in self.topics:
                topic_info = indexQuery(id=topic_id)
                if topic_info: r.append({ 'id': topic_id, 'title': topic_info[0].title })

        return tuple(r)

    def super_topics(self):
        """The list of all topics this document implicitly resides in including
           all parent topics up to but not including the root"""

        r = {}  # Use a dict to filter out duplicates
        if hasattr(self, 'topics') and hasattr(self.aq_parent, 'Index'):
            indexQuery = self.aq_parent.Index.query # Acquire the index's query method

            # Iterate the topic ids assigned to this document and
            # search the topic index to find the parents of these topics
            for topic_id in self.topics:
                for topic_info in indexQuery(id=topic_id):
                    topic = topic_info.getObject()
                
                    if topic:
                        for parent in topic.topicParents()[1:]:
                            r[parent.id] = 1
        return r.keys()
                    

    def all_searchable_text(self):
        """All searchable text data concatenated together"""

        r = []

        # Add all string and text properties not marked hidden
        for p in self.propertyMap():
            if p.get('type', '') in ['string', 'text'] and not p.get('hidden', 0):
                v = self.getProperty(p['id'])
                if v: r.append(v)

        # Add the titles of the topics assigned to this document
        for t in self.topicMap():
            r.append(t['title'])
        
        # Add the full text content of the document
        v = self.PrincipiaSearchSource()
        if v: r.append(v)

        return join(r, '\n')

    def owner_id(self):
        """The user id of the user who owns this document"""
        owner = self.owner_info()
        if owner and owner.has_key('id'):
            return owner['id']
        else:
            return None           

    def authenticated_user_is_owner(self, REQUEST=None):
        """Returns true if the currently authenticated user owns
           the document object"""
        if not REQUEST: REQUEST = self.REQUEST
        return self.owner_id() == REQUEST.AUTHENTICATED_USER.getUserName()

    #
    #   Dublin core compliance methods
    #

    creator = '' # Override Zope creator method

    def format(self):
        """Return the MIME content-type of the document"""
        if hasattr(self, 'aq_base'): self = self.aq_base
        c_type = getattr(self, 'content_type', None)
        if c_type: c_type = split(c_type, ';')[0]
        return c_type

    def source(self):
        """Return the URI of the source document if it was
           submitted via URI"""
        if hasattr(self, 'aq_base'): self = self.aq_base
        return getattr(self, 'data_url', None)

    def identifier(self):
        """Unambiguous document identifier"""
        if not hasattr(self, 'aq_base'): return self.getId()
        if hasattr(self.aq_base, 'filename'):
            return self.absolute_url() + '/' + urllib.quote(self.filename)
        else:
            return self.absolute_url() 

    def subject(self):
        """Return the subject as a comma separated list of
           the topic titles this document resides in"""
        r = map(lambda i: i['title'], self.topicMap())
        return join(r, ', ')

    #
    #   Property handling methods
    #

    def _getDocumentProperties(self):
        """Acquire the property information from the document store"""
        try:
            if self._properties != self.aq_parent._documentProperties:
                self._properties = self.aq_parent._documentProperties
        except:
            pass

        return self._properties

    def propdict(self):
        dict={}
        for p in self._getDocumentProperties():
            dict[p['id']]=p
        return dict
    
    def propertyMap(self):
        """Override the default property mapping for File objects.
           Documents use a single property definition in the store instead"""

        # Return a tuple of mappings, giving meta-data for document properties.
        r=[]
        for d in self._getDocumentProperties():
            mode=d.get('mode', 'wd')
            if 'd' in mode:
                dd={}
                dd.update(d)
                d=dd
                d['mode']=filter(lambda c: c != 'd', mode)
            r.append(d)
            
        return tuple(r)

    def hasProperty(self, id):
        """Return true if object has a property 'id'"""
        for p in self._getDocumentProperties():
            if id==p['id']:
                return 1
        return 0

    def property_extensible_schema__(self):
        """Document files have fixed property schemas if they are in
           a document store which describes their properties"""
        return not hasattr(self.aq_parent, '_documentProperties')

    def manage_edit(self, title, content_type, precondition='', REQUEST=None):
        """Update object and reindex"""
        r = File.manage_edit(self, title, content_type, precondition, REQUEST)
        self.reindex_object()
        return r

    def _updateProperty(self, name, value):
        """Update property value by name. Includes special parsing for
           topic list properties"""

        if name == 'topics':
            # Filter the topic code list to get the topic codes alone
            topics = []
            for topic in value:
                if strip(topic): topics.append(split(topic)[0])
            r = File._updateProperty(self, name, topics)
            if hasattr(self.aq_base, 'review_date') and self.review_date != DLGlobals.document_unreviewed_date:
                self.showTopics()
        else:
            r = File._updateProperty(self, name, value)
    
    def manage_changeProperties(self, REQUEST=None, reindex=1, **kw):
        """Update properties and reindex"""
        r = File.manage_changeProperties(self, REQUEST, kw=kw)
        if reindex: self.reindex_object()
        return r

    def manage_editProperties(self, REQUEST, reindex=1):
        """Edit Properties and reindex"""
        r = File.manage_editProperties(self, REQUEST)
        if reindex: self.reindex_object() 
        return r

    #
    # Document data handlers
    #

    def update_data(self, data, content_type=None, size=None):
        """Update file data and reindex"""
        r = File.update_data(self, data, content_type, size)
        self.setIcons()
        # Clear any cached file data
        if hasattr(self, '_v_raw_text'): del self._v_raw_text
        # reindex the document in the Catalog
        self.reindex_object()
        return r

    def manage_upload(self, file, REQUEST=None, remote=0):
        """Upload new file data into the document. Extended to support
           uploading data from remote servers"""
        
        if remote and type(file) == type(''):
            url = file

            # make sure the url uses an acceptable protocol (not file or mailto)
            p = find(url, ':')
            if p > -1:
                if url[:p] not in ('http', 'ftp', 'gopher'):
                    raise "URLError", \
                          ('<span>The URL specified: '
                           '<a href="%s" target="_blank">%s</a> '
                           'is invalid.</span>') % (url, url)
            else:
                # use HTTP if no protocol was specified
                url = 'http://' + url

            self.data_url = url          
            # Attempt to connect and retrieve the remote document
            try:
                remote_file = urllib.urlopen(url)
                data = remote_file.read()
                if remote_file.headers.type == 'text/html':
                    data = unlocalizeHTML(url, data)
                file = Pdata(data) # blob wrapper object from OFS.Image
                file.headers = remote_file.info().dict
                # get the filename from the url
                p = rfind(remote_file.url, '/')
                if p == len(remote_file.url) - 1 or remote_file.url[p-2:p+1] == '://':
                    file.filename = 'index.html'
                elif p > -1:
                    file.filename = remote_file.url[p+1:]
                
                remote_file.close()
            except IOError, error:
                raise "URLError", \
                      ('<div>An error occurred while '
                       'attempting to connect to '
                       '<a href="%s" target="_blank">%s</a></div>' 
                       '<div>Error Message: <em>%s</em></div>') % (url, url, error)
        else:
            if hasattr(self.aq_base, 'data_url'): del self.data_url

        # Set the filename
        filename = getattr(file, 'filename', 'untitled')

        # Strip off the file path (if any)
        for path_sep in ['\\', '/', ':']:
            sep = rfind(filename, path_sep) + 1
            if sep: 
                filename = filename[sep:]
                break

        while hasattr(self, filename):
            # Make sure the filename is not the same name as existing attrs
            filename = 'doc_' + filename
        
        self.filename = filename

        # Upload the data
        return File.manage_upload(self, file, REQUEST)

    def setIcons(self):
        """set the document icons based on the content-type"""
        format = self.format()
        sep = find(format,'/') + 1
        type = format[sep:]
        sep = find(type, '-') + 1
        if sep: type = type[sep:]
        large_icon_path = 'document_icons/%s_large.gif' % type
        if path.exists(path.join(package_home(globals()), large_icon_path)):
            self.large_icon = ImageFile(large_icon_path, globals())
        else:
            self.large_icon = ImageFile('document_icons/generic_large.gif', globals())

        small_icon_path = 'document_icons/%s_small.gif' % type
        if path.exists(path.join(package_home(globals()), small_icon_path)):
            self.small_icon = ImageFile(small_icon_path, globals())
        else:
            self.small_icon = ImageFile('document_icons/generic_small.gif', globals())

        self.icon = join(self.getPhysicalPath()[1:],'/') + '/small_icon'

    def showTopics(self):
        """Make the document topics visible"""
        for topic_id in self.topics:
            topics = self.aq_parent.Index.query(id=topic_id)
            for topic in topics:
                topic = topic.getObject()
                if topic: topic.showTopic() # Sometime getObject() returns None, gaack!


Globals.InitializeClass(Document)

class DocURLopener(urllib.FancyURLopener):
    """Extend the urllib FancyURLopener to raise an exception on 404 not found errors."""
    
    def http_error_404(self, url, fp, errcode, errmsg, headers):
        raise "NotFoundError", \
              ('<div>The document specified at: '
               '<a href="http:%s" target="_blank">http:%s</a> '
               'was not found. ' 
               'Please double check this URL and '
               'submit again.</div>') % (url, url)

urllib._urlopener = DocURLopener() # Tell urllib to use our urlopener class

def unlocalizeHTML(url, data):
    """Add a <base> declaration to documents that lack it to compensate for 
       relocating them into the library"""

    p = find(data, '<base')
    if p == -1: p = find(data, '<BASE')
    if p == -1:
        p = rfind(url, '/')
        if url[p-2:p+1] != '://':
            base_url = url[:p+1]
        else:
            base_url = url + '/'

        p = find(data, '<head>')
        if p > 0: 
            return replace(data, '<head>', '<head><base href="%s">' % base_url, 1)
        elif p == -1:
            p = find(data, '<HEAD>')
            if p > 0:
                return replace(data, '<HEAD>', '<HEAD><BASE href="%s">' % base_url, 1)
    return data
